#include <assert.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "expenv.h"

void *expenv_construct(int sr, double amp, double att, double dec,
                       double sus, double suslvl, double rel, double curve)
{
    EXPENV_DATA *data = (EXPENV_DATA *) malloc(sizeof(EXPENV_DATA));

    assert(curve != 0);
    data->sr = sr;
    data->amp = amp;
    expenv_set_param(data, EXPENV_ATT, att);
    expenv_set_param(data, EXPENV_DEC, dec);
    expenv_set_param(data, EXPENV_SUS, sus);
    expenv_set_param(data, EXPENV_SUSLVL, suslvl);
    expenv_set_param(data, EXPENV_REL, rel);
    expenv_set_param(data, EXPENV_CURVE, curve);

    data->counter = data->samples[0];
    data->stage = 0;
    return (void *) data;
}


void expenv_set_param(void *data, int index, double value)
{
    EXPENV_DATA *d = (EXPENV_DATA *) data;

    switch(index) {
    case EXPENV_AMP:
        d->amp = value;
        break;
    case EXPENV_ATT:
        d->samples[0] = (int) (value * d->sr);

        break;
    case EXPENV_DEC:
        d->samples[1] = (int) (value * d->sr);

        break;
    case EXPENV_SUS:
        d->samples[2] = (int) (value * d->sr);

        break;
    case EXPENV_SUSLVL:
        d->suslvl = value;

        break;
    case EXPENV_REL:
        d->samples[3] = (int) (value * d->sr);

        break;
    case EXPENV_CURVE:
        d->curve = value;

        break;
    }

}

void expenv_process(void *data)
{
    EXPENV_DATA *d = (EXPENV_DATA *) data;
    double out;
    switch (d->stage) {
    case 0:
        out = d->amp * pow((d->samples[d->stage] - d->counter)/(float)d->samples[d->stage], d->curve);
        break;
    case 1:
        out = d->amp
                - ((d->amp - d->suslvl) *
                   pow((d->samples[d->stage] - d->counter)/(float)d->samples[d->stage], 1.0/d->curve));
        break;
    case 2:
        out = d->suslvl;
        break;
    case 3:
        out = d->suslvl * (1 - pow((d->samples[d->stage] - d->counter)/(float)d->samples[d->stage], d->curve));
        break;
    default:
        out = 0.0;
        break;

    }

    d->counter--;

    if(d->counter == 0 && d->stage < 4) {
        d->counter = d->samples[++d->stage];
    }

    *d->out = out;
}


void expenv_destroy(void *data)
{
    EXPENV_DATA *d = (EXPENV_DATA *) data;
    free(d);
}

double *expenv_get_input(void *data, int index)
{
    EXPENV_DATA *d = (EXPENV_DATA *) data;
    switch (index) {
    case EXPENV_TRIGGER:
        return d->trigger;
    }
    return NULL;
}


void expenv_set_output(void *data, int index, double *out_buffer)
{
    EXPENV_DATA *d = (EXPENV_DATA *) data;
    switch (index) {
    case 0:
        d->out = out_buffer;
        break;
    }
}

void expenv_get_ugen_info(ugen_t *ugen_data)
{
    strcpy(ugen_data->name, "expenv");
    ugen_data->process = expenv_process;
    ugen_data->destroy = expenv_destroy;
    ugen_data->set_param = expenv_set_param;
    ugen_data->get_input = expenv_get_input;
    ugen_data->set_output = expenv_set_output;
}
