#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "graph.h"
#include "expenv.h"
#include "linen.h"
#include "sinosc.h"
#include "wshaper.h"

void graph_ugen_free(graph_node_t *node);

graph_t *graph_new(int inchannels, int outchannels)
{
    graph_t *graph = (graph_t *) malloc(sizeof(graph_t));
    graph->inchannels = inchannels;
    graph->outchannels = outchannels;
    if (graph->inchannels != 0) {
        graph->inbuf = (double *) calloc(graph->inchannels, sizeof(double));
    } else {
        graph->inbuf = NULL;
    }
    if (graph->outchannels != 0) {
        graph->outbuf = (double *) calloc(graph->outchannels, sizeof(double));
    } else {
        graph->outbuf = NULL;
    }

    graph->head = (graph_node_t *) malloc(sizeof(graph_node_t));
    graph->head->next = NULL;
    graph->head->ugen_instance = NULL;

    graph->ugens = (ugen_t *) malloc(sizeof(ugen_t));
    ugen_t *ugen = graph->ugens;
    expenv_get_ugen_info(ugen);

    ugen->next = (ugen_t *) malloc(sizeof(ugen_t));
    ugen = ugen->next;
    linen_get_ugen_info(ugen);

    ugen->next = (ugen_t *) malloc(sizeof(ugen_t));
    ugen = ugen->next;
    sinosc_get_ugen_info(ugen);

    ugen->next = (ugen_t *) malloc(sizeof(ugen_t));
    ugen = ugen->next;
    wshaper_get_ugen_info(ugen);

    ugen->next = NULL;

    return graph;
}


void graph_delete(graph_t *graph)
{
    if (graph->inbuf) {
        free(graph->inbuf);
    }
    if (graph->outbuf) {
        free(graph->outbuf);
    }

    graph_node_t *node = graph->head->next;
    while (node) {
        graph_ugen_free(node);
        node = node->next;
    }

    free(graph->head);
    free(graph);
}


graph_node_t *graph_create_node(graph_t *graph, const char *name, ...)
{
    graph_node_t *node;
    va_list ap;
    ugen_t* ugen = graph->ugens;

    while(ugen) {
        if (strncmp(ugen->name, name, 32) == 0) {
            break;
        }
        ugen = ugen->next;
    }
    if(!ugen) {
        return NULL;
    }

    node = (graph_node_t *) malloc(sizeof(graph_node_t));
    if (strcmp(name, "linen") == 0) {
        va_start(ap, 5);
        int sr = va_arg(ap, int);
        double amp = va_arg(ap, double);
        double att = va_arg(ap, double);
        double sus = va_arg(ap, double);
        double rel = va_arg(ap, double);
        va_end(ap);

        node->ugen_instance = (void *) linen_construct(sr, amp, att, sus, rel);
        node->ugen = ugen;

    } else if (strcmp(name, "wshaper") == 0) {

        node->ugen_instance = (void *) wshaper_construct();
        node->ugen = ugen;
    } else if (strcmp(name, "sinosc") == 0) {
        va_start(ap, 3);
        int sr = va_arg(ap, int);
        double freq = va_arg(ap, double);
        double amp = va_arg(ap, double);
        va_end(ap);

        node->ugen_instance = (void *) sinosc_construct(sr, freq, amp);
        node->ugen = ugen;
    }

    graph_node_t *last = graph->head;
    while (last->next) {
        last = last->next;
    }
    last->next = node;

    return node;
}

void graph_ugen_free(graph_node_t *node)
{
    // FIXME free graph
}


void graph_connect(graph_t *graph, graph_node_t *src, int srcindex, graph_node_t *dst, int dstindex)
{
    // TODO validate that nodes are in graph;
    src->ugen->set_output(src->ugen_instance, srcindex,
                          dst->ugen->get_input(dst->ugen_instance, dstindex));
}


void graph_set_out(graph_t *graph, graph_node_t *src, int srcindex, int dstindex)
{
    // TODO validate srcindex and dstindex
    src->ugen->set_output(src->ugen_instance, srcindex, graph->outbuf + dstindex);
}


void graph_put_input(graph_t *graph, double *in)
{
    // TODO implement input mapping
    memcpy(graph->inbuf, in, graph->inchannels * sizeof(double));
}


double *graph_get_next(graph_t *graph)
{
    graph_node_t *node = graph->head->next;
    while(node) {
        node->ugen->process(node->ugen_instance);
        node = node->next;
    }
    return graph->outbuf;
}
